@@ -1,5 +1,8 @@
 package XPAN::Query;
 
+our $DATE = '2014-12-14'; # DATE
+our $VERSION = '0.06'; # VERSION
+
 use 5.010001;
 use strict;
 use warnings;
@@ -7,8 +10,7 @@ use Log::Any '$log';
 
 use Digest::MD5 qw(md5_hex);
 use File::Slurp::Tiny qw(read_file write_file);
-use LWP::UserAgent;
-use Sereal qw(encode_sereal decode_sereal);
+use PerlIO::gzip;
 use String::ShellQuote;
 use URI;
 
@@ -22,24 +24,35 @@ our @EXPORT_OK = qw(
                );
 
 our %SPEC;
-our $VERSION = '0.05'; # VERSION
-our $DATE = '2014-08-16'; # DATE
+
+our $CACHE_PERIOD = $ENV{XPAN_CACHE_PERIOD} // 86400;
+our $URL          = $ENV{XPAN_URL} // ["/cpan", "http://www.cpan.org/"];
 
 my %common_args = (
     url => {
         summary => "URL to repository, e.g. '/cpan' or 'http://host/cpan'",
-        schema  => 'str*',
-        req => 1,
-        pos => 0,
+        schema  => [str => default => $URL],
+        description => <<'_',
+
+If not specified, will default to `XPAN_URL` environment, or `$URL` variable
+(which by default is set to `/cpan`).
+
+_
     },
     cache_period => {
-        schema => [int => default => 86400],
+        schema => [int => default => $CACHE_PERIOD],
         cmdline_aliases => {
             nocache => {
                 schema => [bool => {is=>1}],
                 code   => sub { $_[0]{cache_period} = 0 },
             },
         },
+        description => <<'_',
+
+If you set this to 0 it means to force cache to expire. If you set this to -1 it
+means to never expire the cache (always use the cache no matter how old it is).
+
+_
     },
     detail => {
         summary => "If set to true, will return array of records instead of just ID's",
@@ -55,7 +68,7 @@ my %query_args = (
         summary => 'Search query',
         schema => 'str*',
         cmdline_aliases => {q=>{}},
-        pos => 1,
+        pos => 0,
     },
 );
 
@@ -64,45 +77,78 @@ sub _parse {
     my %args = @_;
 
     my $now = time();
-
-    my $xpan_url = $args{url} or die "Please supply url";
-    # normalize for LWP, it won't accept /foo/bar, only file:/foo/bar
-    $xpan_url = URI->new($xpan_url);
-    unless ($xpan_url->scheme) { $xpan_url = URI->new("file:$args{url}") }
-
     my $tmpdir = $args{temp_dir} // $ENV{TEMP} // $ENV{TMP} // "/tmp";
-    my $cache_period = $args{cache_period} // 86400;
-
-    state $ua = LWP::UserAgent->new;
+    my $cache_period = $args{cache_period} // $CACHE_PERIOD;
+    state $ua = do { require LWP::UserAgent; LWP::UserAgent->new };
+    my $url0 = $args{url} // $URL or die "Please supply url";
     my $filename = "02packages.details.txt";
-    my $md5 = md5_hex("$xpan_url");
 
-    # download file
-    my $gztarget = "$tmpdir/$filename.gz-$md5";
-    my @gzst = stat($gztarget);
-    if (@gzst && $gzst[9] >= $now-$cache_period) {
-        $log->tracef("Using cache file %s", $gztarget);
-    } else {
-        my $url = "$xpan_url/modules/$filename.gz";
-        $log->tracef("Downloading %s ...", "$url");
-        my $res = $ua->get($url);
-        unless ($res->is_success) {
-            die "Can't get $url: " . $res->status_line;
+    my $has_success_url;
+    my $md5;
+    my @gzst;
+    my $gztarget;
+
+  DOWNLOAD:
+    for my $xpan_url0 (ref($url0) eq 'ARRAY' ? @$url0 : $url0) {
+
+        # normalize for LWP, it won't accept /foo/bar, only file:/foo/bar
+        my $xpan_url = URI->new($xpan_url0);
+        unless ($xpan_url->scheme) { $xpan_url = URI->new("file:$xpan_url0") }
+
+        $md5 = md5_hex("$xpan_url");
+
+        # download file
+        $gztarget = "$tmpdir/$filename.gz-$md5";
+        @gzst = stat($gztarget);
+        if (@gzst && ($cache_period < 0 || $gzst[9] >= $now-$cache_period)) {
+            $log->tracef("Using cached download file %s", $gztarget);
+            $has_success_url++;
+            last DOWNLOAD;
+        } else {
+            my $url = "$xpan_url/modules/$filename.gz";
+            $log->tracef("Downloading %s ...", $url);
+            my $res = $ua->get($url);
+            unless ($res->is_success) {
+                $log->warnf("Can't get %s: %s", $url, $res->status_line);
+                next DOWNLOAD;
+            }
+            $has_success_url++;
+            $log->tracef("Writing %s ...", $gztarget);
+            write_file($gztarget, $res->content);
         }
-        write_file($gztarget, $res->content);
     }
 
-    # extract and process (XXX this is currently unix-specific)
-    my $sertarget = "$tmpdir/$filename.2.sereal-$md5";
-    my @serst = stat($sertarget);
-    my $data;
-    if (@serst && $serst[9] >= $gzst[9]) {
-        $log->tracef("Using cache file %s", $sertarget);
-        $data = decode_sereal(~~read_file($sertarget));
+    die "No mirrors available" unless $has_success_url;
+
+    # extract and convert to SQLite database
+
+    require DBI;
+    my $sqlitetarget = "$tmpdir/$filename.sqlite-$md5";
+    my @sqst = stat($sqlitetarget);
+    my $dbh;
+    if (@sqst && $sqst[9] >= $gzst[9]) {
+        $log->tracef("Using cached SQLite file %s", $sqlitetarget);
+        $dbh = DBI->connect("dbi:SQLite:dbname=$sqlitetarget", undef, undef,
+                            {RaiseError=>1});
     } else {
-        $log->trace("Parsing $filename.gz ...");
-        my (%packages, %authors, %dists);
-        open my($fh), "zcat ".shell_quote("$gztarget")."|";
+        $log->tracef("Creating %s ...", $sqlitetarget);
+
+        require IO::Compress::Gzip;
+
+        open my($fh), "<:gzip", $gztarget
+            or die "Can't open $gztarget (<:gzip): $!";
+
+        unlink $sqlitetarget;
+        $dbh = DBI->connect("dbi:SQLite:dbname=$sqlitetarget", undef, undef,
+                            {RaiseError=>1});
+        $dbh->do("CREATE TABLE author (id TEXT NOT NULL PRIMARY KEY)");
+        $dbh->do("CREATE TABLE dist (name TEXT NOT NULL PRIMARY KEY, author TEXT, version TEXT, file TEXT)");
+        $dbh->do("CREATE INDEX dist_author ON dist(author)");
+        $dbh->do("CREATE TABLE package (name TEXT NOT NULL PRIMARY KEY, author TEXT, version TEXT, file TEXT, dist TEXT)");
+        $dbh->do("CREATE INDEX package_author ON package(author)");
+        $dbh->do("CREATE INDEX package_dist ON package(dist)");
+
+        $dbh->begin_work;
         my $line = 0;
         while (<$fh>) {
             $line++;
@@ -114,28 +160,25 @@ sub _parse {
             $ver = undef if $ver eq 'undef';
             my ($author, $file) = $path =~ m!^./../(.+?)/(.+)!
                 or die "Line $line: Invalid path $path";
-            $authors{$author} = 1;
-            $packages{$pkg} = {author=>$author, version=>$ver, file=>$file};
+            $dbh->do("INSERT OR IGNORE INTO author (id) VALUES (?)", {}, $author);
             my $dist = $file;
             # XXX should've extract metadata
             if ($dist =~ s/-v?(\d(?:\d*(\.[\d_][^.]*)*?)?).\D.+//) {
                 #say "D:  dist=$dist, 1=$1";
-                $dists{$dist} = {author=>$author, version=>$1, file=>$file};
-                $packages{$pkg}{dist} = $dist;
+                $dbh->do("INSERT OR IGNORE INTO dist (name, author, version, file) VALUES (?,?,?,?)", {},
+                         $dist, $author, $1, $file);
+                $dbh->do("INSERT OR IGNORE INTO package (name, author, version, file, dist) VALUES (?,?,?,?,?)", {},
+                         $pkg, $author, $ver, $file, $dist);
             } else {
                 $log->info("Line $line: Can't parse dist version from filename $file");
+                $dbh->do("INSERT OR IGNORE INTO package (name, author, version, file) VALUES (?,?,?,?)", {},
+                         $pkg, $author, $ver, $file);
                 #next;
             }
         }
-        $data = {
-            packages => \%packages,
-            authors  => [sort keys %authors],
-            dists    => \%dists,
-        };
-        write_file($sertarget, encode_sereal($data));
+        $dbh->commit;
     }
-
-    $data;
+    $dbh;
 }
 
 $SPEC{list_xpan_authors} = {
@@ -157,12 +200,12 @@ _
     examples => [
         {
             summary => 'List all authors',
-            argv    => ['/cpan'],
+            argv    => [],
             test    => 0,
         },
         {
-            summary => 'Grep by CPAN ID',
-            argv    => ['/cpan', 'MICHAEL'],
+            summary => 'Find CPAN IDs which start with something',
+            argv    => ['--url', 'http://www.cpan.org/', 'MICHAEL%'],
             result  => ['MICHAEL', 'MICHAELW'],
             test    => 0,
         },
@@ -171,12 +214,25 @@ _
 sub list_xpan_authors {
     my %args = @_;
     my $detail = $args{detail};
-    my $data = _parse(%args);
-    my $q = lc($args{query} // '');
+
+    my $dbh = _parse(%args);
+    my $q = $args{query} // ''; # sqlite is case-insensitive by default, yay
+
+    my @bind;
+    my @where;
+    if (length($q)) {
+        push @where, "(id LIKE ?)";
+        push @bind, $q;
+    }
+    my $sql = "SELECT * FROM author".
+        (@where ? " WHERE ".join(" AND ", @where) : "").
+            " ORDER BY id";
+
     my @res;
-    for (@{ $data->{authors} }) {
-        next if length($q) && index(lc($_), $q) < 0;
-        push @res, $detail ? {cpanid=>$_} : $_;
+    my $sth = $dbh->prepare($sql);
+    $sth->execute(@bind);
+    while (my $row = $sth->fetchrow_hashref) {
+        push @res, $detail ? $row : $row->{id};
     }
     \@res;
 }
@@ -212,16 +268,32 @@ sub list_xpan_packages {
     my %args = @_;
     my $detail = $args{detail};
 
-    my $data = _parse(%args);
-    my $q = lc($args{query} // '');
+    my $dbh = _parse(%args);
+    my $q = $args{query} // ''; # sqlite is case-insensitive by default, yay
+
+    my @bind;
+    my @where;
+    if (length($q)) {
+        push @where, "(name LIKE ?)";
+        push @bind, $q;
+    }
+    if ($args{author}) {
+        push @where, "(author=?)";
+        push @bind, $args{author};
+    }
+    if ($args{dist}) {
+        push @where, "(dist=?)";
+        push @bind, $args{dist};
+    }
+    my $sql = "SELECT * FROM package".
+        (@where ? " WHERE ".join(" AND ", @where) : "").
+            " ORDER BY name";
+
     my @res;
-    for (keys %{ $data->{packages} }) {
-        my $rec = $data->{packages}{$_};
-        next if length($q) && index(lc($_), $q) < 0;
-        next if $args{author} && uc($args{author}) ne uc($rec->{author});
-        next if $args{dist} && $args{dist} ne $rec->{dist};
-        $rec->{name} = $_;
-        push @res, $detail ? $rec : $_;
+    my $sth = $dbh->prepare($sql);
+    $sth->execute(@bind);
+    while (my $row = $sth->fetchrow_hashref) {
+        push @res, $detail ? $row : $row->{name};
     }
     \@res;
 }
@@ -264,12 +336,12 @@ _
     examples => [
         {
             summary => 'List all distributions',
-            argv    => ['/cpan'],
+            argv    => [],
             test    => 0,
         },
         {
             summary => 'Grep by distribution name, return detailed record',
-            argv    => ['/cpan', 'data-table'],
+            argv    => ['--url', '/cpan', 'data-table'],
             result  => [
                 {
                     author  => "BIGJ",                          # ..{0}
@@ -294,7 +366,7 @@ _
         },
         {
             summary   => 'Filter by author, return JSON',
-            src       => 'list-xpan-packages /cpan --author sharyanto --json',
+            src       => 'list-xpan-dists --author perlancar --json',
             src_plang => 'bash',
             test      => 0,
         },
@@ -304,15 +376,28 @@ sub list_xpan_dists {
     my %args = @_;
     my $detail = $args{detail};
 
-    my $data = _parse(%args);
-    my $q = lc($args{query} // '');
+    my $dbh = _parse(%args);
+    my $q = $args{query} // '';
+
+    my @bind;
+    my @where;
+    if (length($q)) {
+        push @where, "(name LIKE ?)";
+        push @bind, $q;
+    }
+    if ($args{author}) {
+        push @where, "(author=?)";
+        push @bind, $args{author};
+    }
+    my $sql = "SELECT * FROM package".
+        (@where ? " WHERE ".join(" AND ", @where) : "").
+            " ORDER BY name";
+
     my @res;
-    for (keys %{ $data->{dists} }) {
-        my $rec = $data->{dists}{$_};
-        next if length($q) && index(lc($_), $q) < 0;
-        next if $args{author} && uc($args{author}) ne uc($rec->{author});
-        $rec->{name} = $_;
-        push @res, $detail ? $rec : $_;
+    my $sth = $dbh->prepare($sql);
+    $sth->execute(@bind);
+    while (my $row = $sth->fetchrow_hashref) {
+        push @res, $detail ? $row : $row->{name};
     }
     \@res;
 }
@@ -333,7 +418,7 @@ XPAN::Query - Query a {CPAN,MiniCPAN,DarkPAN} mirror
 
 =head1 VERSION
 
-This document describes version 0.05 of XPAN::Query (from Perl distribution XPAN-Query), released on 2014-08-16.
+This document describes version 0.06 of XPAN::Query (from Perl distribution XPAN-Query), released on 2014-12-14.
 
 =head1 SYNOPSIS
 
@@ -343,8 +428,19 @@ This document describes version 0.05 of XPAN::Query (from Perl distribution XPAN
      list_xpan_dists
      list_xpan_authors
  );
- my $res = list_ubuntu_releases(detail=>1);
- # raw data is in $Ubuntu::Releases::data;
+
+ # the first query will download 02packages.details.txt.gz from a CPAN mirror
+ # (the default is "/cpan" or "http://www.cpan.org/") and convert it to a SQLite
+ # database, so it will take some time, e.g. several seconds for download (1.5MB
+ # at the time of this writing, so a few seconds depending on your connection
+ # speed) plus around 10-15s for conversion.
+
+ my $res = list_xpan_authors("MICHAEL%"); # => ["MICHAEL", "MICHAELW"]
+
+ # the subsequent queries will be instantaneous, unless you change mirror site
+ # or 24 hours has passed, which is the default cache period.
+
+ my list_xpan_modules(author=>"NEILB", detail=>1);
 
 =head1 DESCRIPTION
 
@@ -368,16 +464,16 @@ List authors in {CPAN,MiniCPAN,DarkPAN} mirror.
 
 Examples:
 
- list_xpan_authors( url => "/cpan");
+ list_xpan_authors();
 
 
 List all authors.
 
 
- list_xpan_authors( query => "MICHAEL", url => "/cpan"); # -> ["MICHAEL", "MICHAELW"]
+ list_xpan_authors( query => "MICHAEL%", url => "http://www.cpan.org/"); # -> ["MICHAEL", "MICHAELW"]
 
 
-Grep by CPAN ID.
+Find CPAN IDs which start with something.
 
 
 Arguments ('*' denotes required arguments):
@@ -386,6 +482,9 @@ Arguments ('*' denotes required arguments):
 
 =item * B<cache_period> => I<int> (default: 86400)
 
+If you set this to 0 it means to force cache to expire. If you set this to -1 it
+means to never expire the cache (always use the cache no matter how old it is).
+
 =item * B<detail> => I<bool>
 
 If set to true, will return array of records instead of just ID's.
@@ -396,10 +495,13 @@ Search query.
 
 =item * B<temp_dir> => I<str>
 
-=item * B<url>* => I<str>
+=item * B<url> => I<str> (default: ["/cpan", "http://www.cpan.org/"])
 
 URL to repository, e.g. '/cpan' or 'http://host/cpan'.
 
+If not specified, will default to C<XPAN_URL> environment, or C<$URL> variable
+(which by default is set to C</cpan>).
+
 =back
 
 Return value:
@@ -416,7 +518,7 @@ List distributions in {CPAN,MiniCPAN,DarkPAN} mirror.
 
 Examples:
 
- list_xpan_dists( url => "/cpan");
+ list_xpan_dists();
 
 
 List all distributions.
@@ -470,6 +572,9 @@ Filter by author.
 
 =item * B<cache_period> => I<int> (default: 86400)
 
+If you set this to 0 it means to force cache to expire. If you set this to -1 it
+means to never expire the cache (always use the cache no matter how old it is).
+
 =item * B<detail> => I<bool>
 
 If set to true, will return array of records instead of just ID's.
@@ -480,10 +585,13 @@ Search query.
 
 =item * B<temp_dir> => I<str>
 
-=item * B<url>* => I<str>
+=item * B<url> => I<str> (default: ["/cpan", "http://www.cpan.org/"])
 
 URL to repository, e.g. '/cpan' or 'http://host/cpan'.
 
+If not specified, will default to C<XPAN_URL> environment, or C<$URL> variable
+(which by default is set to C</cpan>).
+
 =back
 
 Return value:
@@ -508,6 +616,9 @@ Filter by author.
 
 =item * B<cache_period> => I<int> (default: 86400)
 
+If you set this to 0 it means to force cache to expire. If you set this to -1 it
+means to never expire the cache (always use the cache no matter how old it is).
+
 =item * B<detail> => I<bool>
 
 If set to true, will return array of records instead of just ID's.
@@ -522,10 +633,13 @@ Search query.
 
 =item * B<temp_dir> => I<str>
 
-=item * B<url>* => I<str>
+=item * B<url> => I<str> (default: ["/cpan", "http://www.cpan.org/"])
 
 URL to repository, e.g. '/cpan' or 'http://host/cpan'.
 
+If not specified, will default to C<XPAN_URL> environment, or C<$URL> variable
+(which by default is set to C</cpan>).
+
 =back
 
 Return value:
@@ -550,6 +664,9 @@ Filter by author.
 
 =item * B<cache_period> => I<int> (default: 86400)
 
+If you set this to 0 it means to force cache to expire. If you set this to -1 it
+means to never expire the cache (always use the cache no matter how old it is).
+
 =item * B<detail> => I<bool>
 
 If set to true, will return array of records instead of just ID's.
@@ -564,10 +681,13 @@ Search query.
 
 =item * B<temp_dir> => I<str>
 
-=item * B<url>* => I<str>
+=item * B<url> => I<str> (default: ["/cpan", "http://www.cpan.org/"])
 
 URL to repository, e.g. '/cpan' or 'http://host/cpan'.
 
+If not specified, will default to C<XPAN_URL> environment, or C<$URL> variable
+(which by default is set to C</cpan>).
+
 =back
 
 Return value:
@@ -577,6 +697,28 @@ Return value:
 By default will return an array of package names. If you set C<detail> to true,
 will return array of records.
 
+=head1 VARIABLES
+
+=head2 C<$XPAN::Query::CACHE_PERIOD> => int (default: 86400)
+
+Set default cache period, in seconds.
+
+=head2 C<$XPAN::Query::URL> => str (default: "/cpan")
+
+Set default XPAN URL.
+
+=head1 ENVIRONMENT
+
+=head2 XPAN_CACHE_PERIOD => int
+
+Can be used to preset C<$XPAN::Query::CACHE_PERIOD>.
+
+=head2 XPAN_URL => str
+
+Can be used to preset C<$XPAN::Query::URL>.
+
+=head1 TODO
+
 =head1 SEE ALSO
 
 L<Parse::CPAN::Packages> is a more full-featured and full-fledged module to
@@ -612,11 +754,11 @@ feature.
 
 =head1 AUTHOR
 
-Steven Haryanto <stevenharyanto@gmail.com>
+perlancar <perlancar@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2014 by Steven Haryanto.
+This software is copyright (c) 2014 by perlancar@cpan.org.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
